// Copyright Tristan Aubrey-Jones 2014
// For comparison with Flocc auto-generated versions
// see http://www.flocc.net/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace PLINQTests
{
    class Program
    {
        const int mag = 100;
        const int N = 3*mag, M = 3*mag, P = 3*mag;
        static int startTime;
        static bool plinqOn = false;

        static void startTimer(String name)
        {
            Console.Write("Timer starting " + name + "...");
            startTime = Environment.TickCount;
        }

        static void stopTimer(String attr)
        {
            int totalTime = Environment.TickCount-startTime;
            Console.WriteLine("done " + attr + " in " + totalTime + "ms");
        }

        void mm1()
        {
            // setup
            startTimer("mm1");
            var range1 = Enumerable.Range(0, M - 1);
            var range2 = Enumerable.Range(0, P - 1);
            var range3 = Enumerable.Range(0, N - 1);
            if (plinqOn)
            {
                range1 = range1.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
                range2 = range2.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
                range3 = range3.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            }

            // make matrices (b transposed to optimize cache access)
            var A = range1.Select(i => new { ai = i, ajs = range3.Select(j => new { aj=j, av=2.0 }) });
            var B = range2.Select(j => new { bj = j, bis = range3.Select(i => new { bi=i, bv=2.0 }) });

            // multiply them
           // var R1 = A.Join(B, a => a.ai, b => b.bj, (a, b) => new { i=a.ai, j=b.bj, vs=a.ajs.Select(av => av.) });

        }

        static void mm2()
        {
            // setup
            startTimer("mm2");
            var range1 = Enumerable.Range(0, (M * N));
            var range2 = Enumerable.Range(0, (P * N));
            if (plinqOn)
            {
                range1 = range1.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
                range2 = range2.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            }

            // make matrices
            var A = range1.Select(p => new { i = p / N, j = p % N, v = 2.0 });
            var B = range2.Select(p => new { i = p / N, j = p % N, v = 2.0 });

            // multiply them
            var R1 = A.Join(B, a => a.j, b => b.i, (a, b) => new { i=a.i, j=b.j, v=a.v*b.v });
            var R3 = R1.GroupBy(o => new { i = o.i, j = o.j }, o => o.v, (k, vs) => new { i = k.i, j = k.j, v = vs.Sum() });

            // realize value
            //int sz = R3.ToArray().Length; // too big, runs out of memory
            //stopTimer(sz.ToString());

            double[,] res = new double[M, P];
            foreach (var el in R3) res[el.i, el.j] = el.v;
            stopTimer(res[M - 1, P - 1].ToString());
        }

        // from https://gist.github.com/titarenko/3800827
        static void mm3()
        {
            // setup
            startTimer("mm3");

            // make matrices
            double[][] lhs = new double[M][], rhs = new double[P][];
            for (int x = 0; x < M; x++) {
                lhs[x] = new double[N];
                for (int y = 0; y < N; y++) lhs[x][y] = 2.0;
            }
            for (int x = 0; x < P; x++) {
                rhs[x] = new double[N];
                for (int y = 0; y < N; y++) rhs[x][y] = 2.0;
            }

            double[][] res;
            if (plinqOn)
            {
                // multiply in parallel
                res = lhs.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism).Select( // goes through <lhs> row by row
                (row, rowIndex) =>
                rhs[0].AsParallel().Select( // goes through first row of <rhs> cell by cell
                    (_, columnIndex) =>
                    rhs.AsParallel().Select(__ => __[columnIndex]) // selects column from <rhs>
                        .Zip(row.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism), (rowCell, columnCell) => rowCell * columnCell).Sum() // does scalar product
                    ).ToArray()
                ).ToArray(); 
            }
            else
            {
                // multiply sequentially
                res = lhs.Select( // goes through <lhs> row by row
                (row, rowIndex) =>
                rhs[0].Select( // goes through first row of <rhs> cell by cell
                    (_, columnIndex) =>
                    rhs.Select(__ => __[columnIndex]) // selects column from <rhs>
                        .Zip(row, (rowCell, columnCell) => rowCell * columnCell).Sum() // does scalar product
                    ).ToArray()
                ).ToArray();
            }

            // end
            stopTimer(res[M - 1][P - 1].ToString());
        }

        static void mm4()
        {
            // setup
            startTimer("mm4");
            var range1 = Enumerable.Range(0, (M * N));
            var range2 = Enumerable.Range(0, (P * N));
            if (plinqOn)
            {
                range1 = range1.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
                range2 = range2.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            }

            // make matrices
            var A = range1.Select(p => new { i = p / N, j = p % N, v = 2.0 });
            var B = range2.Select(p => new { i = p / N, j = p % N, v = 2.0 });

            // multiply them
            var R1 = from a in A
                     join b in B on a.j equals b.i
                     select new { a=a, b=b };
            var R3 = from ab in R1
                     group ab by new { i = ab.a.i, j = ab.b.j } into g
                     select new { ij=g.Key, v=g.Sum(x => x.a.v * x.b.v) };

            // realize value
            //int sz = R3.ToArray().Length; // too big, runs out of memory
            //stopTimer(sz.ToString());

            double[,] res = new double[M, P];
            foreach (var el in R3) res[el.ij.i, el.ij.j] = el.v;
            stopTimer(res[M - 1, P - 1].ToString());
        }

        //const int histD = 1073741824; // out of memory exception
        const int histD = 1024*1024*32;
        const int histN = 100;
        static void hist1()
        {
            // start time
            startTimer("hist1");

            // create data
            var R = Enumerable.Range(0, histD);
            if (plinqOn)
                R = R.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            Random rng = new Random();
            var D = from r in R select rng.NextDouble();

            // n-bucket histogram
            //double max = D.Max();
            //double min = D.Min();
            var maxmin = D.Aggregate(new { max = Double.MinValue, min = Double.MaxValue }, (acc, v) => new { max=Math.Max(acc.max, v), min=Math.Min(acc.min, v) });
            double max = maxmin.max; double min = maxmin.min;
            double i = (histN - 1.0) / (max - min);
            var D1 = from d in D select (int)(d * i);
            var D2 = from d in D1 group d by d into g select new { v=g.Key, n=g.Count() };
    
            // force value to be realized
            int[] res = new int[histN];
            foreach (var v in D2) res[v.v] = v.n;
            stopTimer(res[0].ToString());
        }

        const int floydN = 400; //2000;
        static void floyd1()
        {
            startTimer("floyd1");

            // init data
            //double[,] I0 = new double[floydN, floydN];
            // TODO ...

            // CAN I USE A MAP, RATHER THAN ENUMERABLE?
            // Can i do the iteration in LINQ, to make it a single statement?

            // do floyd
            var I = Enumerable.Range(0, floydN * floydN).Select(v => new { i = v/floydN, j = v%floydN, v=1.0 });
            if (plinqOn) I = I.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            for (int k = 0; k < floydN; k++)
            {
                var Iik = from i in I where i.i >= 0 && i.i < floydN && i.j == k select i;
                var Ikj = from i in I where i.j >= 0 && i.j < floydN && i.i == k select i;
                var I1 = I.Join(Iik, v => v.i, v => v.i, (ij, ik) => new { ij=ij, ik=ik });
                var I2 = I1.Join(Ikj, v => v.ij.j, v => v.j, (ijik, kj) => new { ijik=ijik, kj = kj });
                I = I2.Select(v => new { i=v.ijik.ij.i, j=v.ijik.ij.j, v=Math.Min(v.ijik.ij.v, v.ijik.ik.v+v.kj.v)} );
                var Iarr = I.ToArray();
                I = Iarr;
                if (plinqOn) I = I.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            }

            double[,] res = new double[floydN, floydN];
            foreach (var v in I) res[v.i, v.j] = v.v;
            stopTimer(res[0,0].ToString());
        }

        // 17:35 - 17:38
        static void dotp()
        {
            startTimer("dotp");

            var A = Enumerable.Range(0, 200000000).Select(v => 1.0);
            var B = Enumerable.Range(0, 100000000).Select(v => 2.0);
            if (plinqOn)
            {
                A = A.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
                B = B.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            }
            var AB = A.Zip(B, (a, b) => a * b);
            var R = A.Sum();

            stopTimer(R.ToString());
        }

        static void tri1()
        {
        }

        // 31/03/2014 11:23 - 11:28 (5mins)
        static void stddev1()
        {
            startTimer("stddev1");
            var X = Enumerable.Range(0, 400000000).Select(v => (double)v);//(v => 2.5);
            if (plinqOn) X = X.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            double num = X.Count();
            double avgX = X.Sum() / num;
            var L1 = X.Select(x => x - avgX).Select(dx => dx*dx);
            var ss = L1.Sum();
            double res = ss / (num - 1);
            stopTimer(res.ToString());
        }

        // 32/03/2014 11:31 - 11:41
        static void linreg()
        {
            startTimer("linreg");
            var L = Enumerable.Range(0, 200000000).Select(v => new { x=1.0, y=2.0 });
            if (plinqOn) L = L.AsParallel().WithExecutionMode(ParallelExecutionMode.ForceParallelism);
            double num = L.Count();
            var tots = L.Aggregate((v1, v2) => new { x=v1.x+v2.x, y=v1.y+v2.y});
            double avgX = tots.x / num, avgY = tots.y / num;
            var L1 = L.Select(v => new { x=v.x-avgX, y=v.y-avgX });
            var L2 = L.Select(v => new { t=v.x*v.y, b=v.x*v.x });
            var topbot = L2.Aggregate((v1,v2) => new { t=v1.t+v2.t, b=v1.b+v2.b });
            double beta = topbot.t / topbot.b;
            double alpha = avgY - (avgX * beta);
            stopTimer(alpha.ToString() + " = y + " + beta.ToString() + "x");
        }

        static void Main(string[] args)
        {
            for (int i = 0; i < 5; i++)
            {
                // sequential
                Console.WriteLine("Sequential");
                plinqOn = false;

                // matrix multiplies
                mm2();
                mm3();
                mm4();
                hist1();
                floyd1();
                dotp();
                stddev1();
                linreg();

                // parallel
                Console.WriteLine("Parallel");
                plinqOn = true;

                mm2();
                mm3();
                mm4();
                hist1();
                floyd1();
                dotp();
                stddev1();
                linreg();
            }

            Console.Write("Press return to exit...");
            Console.ReadLine();
        }
    }
}
